# -*- coding: utf-8 -*-
"""
Created on Sun Jun 27 19:43:03 2021

@author: Philipp
"""

'''
This program is used to calculate the leakage rate of the Normally Open valves with metal sealing.
The leakage channel is a rectangular slit, whose width and lenght are given by the sealing geometry.
The channel height is calculated using Perssons theory of contact mechanics (see MA Maier).
The squeezing pressure is given by the blocking pressure and an additional factor (ratio of sealing area to contact area).
'''

import numpy as np
from scipy import integrate
from uncertainties import ufloat
from uncertainties import unumpy as unp
import scipy.constants as sc

''' Piezo Physics and calculation of C_E and C_p'''

def _Cp(Yp_m, poisson, Yp_p, diaphragm_thickness, membrane_radius, piezo_thickness, piezo_radius):  # CP piezo
    '''
    Parameters
    ----------
    Yp_m : ufloat
        elatic modulus of the membrane [Pa]
    poisson : ufloat
        poisson ratio
    Yp_p : ufloat
        elatic modulus of the piezo [Pa]
    diaphragm_thickness : ufloat
         thickness of the membrane [m]
    membrane_radius : ufloat
         radius of the membrane [m]
    piezo_thickness : ufloat
         thickness of the piezo [m]
    piezo_radius : ufloat
         radius of the piezo [m]
    
    Returns
    -------
    C_p : ufloat
        fluidic capacitance [m^3/Pa]

    '''
    F257 = (16 * diaphragm_thickness ** 3 * Yp_m * (
                    diaphragm_thickness ** 4 * Yp_m ** 2 + 2 * diaphragm_thickness * piezo_thickness * (
                    2 * diaphragm_thickness ** 2 + 3 * diaphragm_thickness * piezo_thickness + 2 * piezo_thickness ** 2) * Yp_m * Yp_p + piezo_thickness ** 4 * Yp_p ** 2) * (
                        4 * diaphragm_thickness ** 4 * Yp_m ** 2 * membrane_radius ** 4 +
                        8 * diaphragm_thickness ** 3 * piezo_thickness * Yp_m * Yp_p * membrane_radius ** 2 * (
                                    membrane_radius - piezo_radius) * (membrane_radius + piezo_radius) * (1 + poisson) +
                        12 * diaphragm_thickness ** 2 * piezo_thickness ** 2 * Yp_m * Yp_p * membrane_radius ** 2 * (
                                    membrane_radius - piezo_radius) * (membrane_radius + piezo_radius) * (1 + poisson) +
                        8 * diaphragm_thickness * piezo_thickness ** 3 * Yp_m * Yp_p * membrane_radius ** 2 * (
                                    membrane_radius - piezo_radius) * (membrane_radius + piezo_radius) * (1 + poisson) +
                        piezo_thickness ** 4 * Yp_p ** 2 * (membrane_radius - piezo_radius) ** 2 * (
                                    membrane_radius + piezo_radius) ** 2 * (1 + poisson) ** 2))

    F258 = 4 * diaphragm_thickness ** 8 * Yp_m ** 4 * membrane_radius ** 10 * (
                    -1 + poisson) + piezo_thickness ** 8 * Yp_p ** 4 * (membrane_radius - piezo_radius) ** 5 * (
                       membrane_radius + piezo_radius) ** 5 * (-1 + poisson) * (1 + poisson) ** 2 + \
               4 * diaphragm_thickness * piezo_thickness ** 7 * Yp_m * Yp_p ** 3 * (
                           membrane_radius - piezo_radius) ** 4 * (membrane_radius + piezo_radius) ** 4 * (
                           -1 + poisson) * (1 + poisson) * (
                       (-piezo_radius) ** 2 * (1 + poisson) + membrane_radius ** 2 * (3 + poisson)) + \
               2 * diaphragm_thickness ** 7 * piezo_thickness * Yp_m ** 3 * Yp_p * membrane_radius ** 4 * (
                           -1 + poisson) * (
                       -13 * membrane_radius ** 4 * piezo_radius ** 2 * (
                           1 + poisson) + 18 * membrane_radius ** 2 * piezo_radius ** 4 * (1 + poisson) + \
                       4 * membrane_radius ** 6 * (3 + poisson) - 3 * piezo_radius ** 6 * (5 + 3 * poisson)) + \
               2 * diaphragm_thickness ** 2 * piezo_thickness ** 6 * Yp_m * Yp_p ** 2 * (
                           membrane_radius - piezo_radius) ** 4 * (membrane_radius + piezo_radius) ** 4 * (
                           -1 + poisson) * (1 + poisson) * (
                       16 * Yp_m * membrane_radius ** 2 + 3 * Yp_p * (
                           (-piezo_radius) ** 2 * (1 + poisson) + membrane_radius ** 2 * (3 + poisson)))

    F259 = diaphragm_thickness ** 4 * piezo_thickness ** 4 * Yp_m ** 2 * Yp_p ** 2 * membrane_radius ** 2 * (
                    membrane_radius - piezo_radius) * (membrane_radius + piezo_radius) * (-1 + poisson) * (
                       -(piezo_radius ** 6) * (1 + poisson) * (131 + 3 * poisson) +
                       2 * membrane_radius ** 2 * piezo_radius ** 4 * (197 + 3 * poisson * (
                           66 + poisson)) - 4 * membrane_radius ** 4 * piezo_radius ** 2 * (
                               96 + poisson * (98 + poisson)) + membrane_radius ** 6 * (
                               141 + poisson * (
                                   138 + poisson))) + 4 * diaphragm_thickness ** 5 * piezo_thickness ** 3 * Yp_m ** 2 * Yp_p * membrane_radius ** 2 * (
                       membrane_radius - piezo_radius) * (membrane_radius + piezo_radius) * (
                       3 * Yp_p * (8 * membrane_radius ** 6 - 21 * membrane_radius ** 4 * piezo_radius ** 2 +
                                   21 * membrane_radius ** 2 * piezo_radius ** 4 - 7 * piezo_radius ** 6) * (
                                   -1 + poisson ** 2) + 2 * Yp_m * membrane_radius ** 2 * (
                               -1 + poisson) * (
                               membrane_radius ** 4 * (3 + poisson) - membrane_radius ** 2 * piezo_radius ** 2 * (
                                   1 + 3 * poisson) + piezo_radius ** 4 * (
                                       5 + 3 * poisson))) + 4 * diaphragm_thickness ** 6 * piezo_thickness ** 2 * Yp_m ** 2 * Yp_p * membrane_radius ** 2 * (
                           membrane_radius - piezo_radius) * (
                       membrane_radius + piezo_radius) * (
                       2 * Yp_p * (4 * membrane_radius ** 6 - 9 * membrane_radius ** 4 * piezo_radius ** 2 +
                                   9 * membrane_radius ** 2 * piezo_radius ** 4 - 3 * piezo_radius ** 6) * (
                                   -1 + poisson ** 2) + 3 * Yp_m * membrane_radius ** 2 * (
                               -1 + poisson) * (
                               membrane_radius ** 4 * (3 + poisson) - membrane_radius ** 2 * piezo_radius ** 2 * (
                                   1 + 3 * poisson) + piezo_radius ** 4 * (5 + 3 * poisson)))

    F260 = diaphragm_thickness ** 3 * piezo_thickness ** 5 * Yp_m * Yp_p ** 2 * (
                    membrane_radius - piezo_radius) ** 2 * (membrane_radius + piezo_radius) ** 2 * (-1 + poisson) * (
                           1 + poisson) * (
                       96 * Yp_m * membrane_radius ** 2 * (membrane_radius - piezo_radius) ** 2 * (
                           membrane_radius + piezo_radius) ** 2 +
                       Yp_p * (-3 * piezo_radius ** 6 * (1 + poisson) + 4 * membrane_radius ** 6 * (
                           3 + poisson) + 4 * membrane_radius ** 2 * piezo_radius ** 4 * (
                                       5 + 3 * poisson) - 2 * membrane_radius ** 4 * piezo_radius ** 2 * (
                                           11 + 6 * poisson)))

    return -(np.pi * (1 + poisson) * (F258 + F259 + F260)) / F257

def _CEs(Yp_m, poisson, Yp_p, d31, diaphragm_thickness, membrane_radius, piezo_thickness, piezo_radius):  # CE*
    '''
    Parameters
    ----------
    Yp_m : ufloat
        elatic modulus of the membrane [Pa]
    poisson : ufloat
        poisson ratio
    Yp_p : ufloat
        elatic modulus of the piezo [Pa]
    d31 : ufloat
        piezo d31 coefficient [C/N]
    diaphragm_thickness : ufloat
         thickness of the membrane [m]
    membrane_radius : ufloat
         radius of the membrane [m]
    piezo_thickness : ufloat
         thickness of the piezo [m]
    piezo_radius : ufloat
         radius of the piezo [m]
    
    Returns
    -------
    C_E : ufloat
        volumetric–electrical coupling coefficient [m^4/V]

    '''

    F262 = 48 * d31 * diaphragm_thickness ** 9 * piezo_thickness * Yp_m ** 4 * Yp_p * membrane_radius ** 4 * piezo_radius ** 2 * (
                    membrane_radius - piezo_radius) * (membrane_radius + piezo_radius) + \
               4 * diaphragm_thickness ** 8 * Yp_m ** 3 * membrane_radius ** 4 * (
                       12 * d31 * piezo_thickness ** 2 * Yp_m * Yp_p * piezo_radius ** 2 * (
                           membrane_radius - piezo_radius) * (membrane_radius + piezo_radius) +
                       48 * d31 * piezo_thickness ** 2 * Yp_p ** 2 * piezo_radius ** 2 * (
                                   membrane_radius - piezo_radius) * (membrane_radius + piezo_radius)) + \
               480 * d31 * diaphragm_thickness ** 7 * piezo_thickness ** 3 * Yp_m ** 3 * Yp_p ** 2 * membrane_radius ** 4 * piezo_radius ** 2 * (
                           membrane_radius - piezo_radius) * (membrane_radius + piezo_radius) + \
               480 * d31 * diaphragm_thickness ** 6 * piezo_thickness ** 4 * Yp_m ** 3 * Yp_p ** 2 * membrane_radius ** 4 * piezo_radius ** 2 * (
                           membrane_radius - piezo_radius) * (membrane_radius + piezo_radius) + \
               4 * diaphragm_thickness ** 5 * piezo_thickness ** 3 * Yp_m ** 2 * Yp_p * membrane_radius ** 2 * (
                           membrane_radius - piezo_radius) * (membrane_radius + piezo_radius) * \
               (
                       48 * d31 * piezo_thickness ** 2 * Yp_m * Yp_p * membrane_radius ** 2 * piezo_radius ** 2 + 12 * d31 * piezo_thickness ** 2 * Yp_p ** 2 * membrane_radius ** 2 * piezo_radius ** 2) + \
               48 * d31 * diaphragm_thickness ** 4 * piezo_thickness ** 6 * Yp_m ** 2 * Yp_p ** 3 * membrane_radius ** 4 * piezo_radius ** 2 * (
                           membrane_radius - piezo_radius) * (membrane_radius + piezo_radius)

    F263 = 16 * diaphragm_thickness ** 3 * Yp_m * (
                diaphragm_thickness ** 4 * Yp_m ** 2 + 2 * diaphragm_thickness * piezo_thickness * Yp_m * Yp_p * (
                    2 * diaphragm_thickness ** 2 + 3 * diaphragm_thickness * piezo_thickness + 2 * piezo_thickness ** 2) +
                piezo_thickness ** 4 * Yp_p ** 2) * (
                           4 * diaphragm_thickness ** 4 * Yp_m ** 2 * membrane_radius ** 4 + 8 * diaphragm_thickness ** 3 * piezo_thickness * Yp_m * Yp_p *
                           (poisson + 1) * membrane_radius ** 2 * (membrane_radius - piezo_radius) * (
                                   membrane_radius + piezo_radius) + 12 * diaphragm_thickness ** 2 * piezo_thickness ** 2 * Yp_m * Yp_p * (
                                       poisson + 1) *
                           membrane_radius ** 2 * (membrane_radius - piezo_radius) * (
                                       membrane_radius + piezo_radius) + 8 * diaphragm_thickness * piezo_thickness ** 3 * Yp_m * Yp_p * (
                                   poisson + 1) * membrane_radius ** 2 * (membrane_radius - piezo_radius) *
                           (membrane_radius + piezo_radius) + piezo_thickness ** 4 * Yp_p ** 2 * (poisson + 1) ** 2 * (
                                       membrane_radius - piezo_radius) ** 2 * (
                                   membrane_radius + piezo_radius) ** 2)

    CE = -np.pi * (poisson + 1) * F262 / F263  # CE
    
    return CE / piezo_thickness
    
# Rechnung Matin Richter
def p_block(U_plus, U_minus, C_E,C_p):
    p_0 = 101325
    p = -C_E/C_p * (U_plus - U_minus) + p_0   # ohne 1/T_p
    return p




''' Contact Mechanics '''

def integrand_alpha(x,H, gamma=1):
    '''
    Parameters
    ----------
    x : integration variable
    H : ufloat
        Hurst exponent [no unit]

    Returns
    -------
    integrand for calculation of the constant alpha in contact mechanics

    '''
    return unp.sqrt(2*H*(1-H)/np.pi) * gamma * x**(-H)*(x**2 - x**(2*H))**(-0.5)

def alpha(q,H, gamma=1):
    '''
    Parameters
    ----------
    q : ufloat
        ratio of the largest to the smallest wave vector q_1/q_0 [no unit]
    H : ufloat
        Hurst exponent [no unit]

    Returns
    -------
    1/alpha : ufloat
        constant beta in contact mechanics

    '''
    alpha = integrate.quad(integrand_alpha, 1,q, args=(H.n))[0] 
    return 1/alpha

def integrand_beta_zähler(x, H):
    '''
    Parameters
    ----------
    x : integration variable
    H : ufloat
        Hurst exponent [no unit]

    Returns
    -------
    integrand in enumerator for calculation of the constant beta in contact mechanics

    '''
    return x**(-H)*(x**2 - x**(2*H))**(-0.5) * unp.log(H/(2*(1-H))*(x**(2-2*H)-1))

def integrand_beta_nenner(x, H):
    '''
    Parameters
    ----------
    x : integration variable
    H : ufloat
        Hurst exponent [no unit]

    Returns
    -------
    integrand in denominator for calculation of the constant beta in contact mechanics

    '''
    return 2 * x**(-H)*(x**2 - x**(2*H))**(-0.5)

def beta(q,H, gamma=1):
    '''
    Parameters
    ----------
    q : ufloat
        ratio of the largest to the smallest wave vector q_1/q_0 [no unit]
    H : ufloat
        Hurst exponent [no unit]

    Returns
    -------
    exp(beta) : ufloat
        constant beta in contact mechanics

    '''
    zähler = integrate.quad(integrand_beta_zähler, 1, q, args=(H.n))[0]
    nenner = 2 * integrate.quad(integrand_beta_nenner, 1, q, args=(H.n))[0]
    beta = zähler/nenner
    return unp.exp(beta)

def average_separation(p, h_rms, q, q_0, H, E, gamma = 1, eps = 0.7493):
    '''
    Parameters
    ----------
    p : ufloat
        squeezing pressure [Pa]
    h_rms : ufloat
        root mean square roughness [m]
    q : ufloat
        ratio of the largest to the smallest wave vector q_1/q_0 [no unit]
    q_0 : ufloat
        smallest wave vector [1/m]
    H : ufloat
        Hurst exponent [no unit]
    E : ufloat
        Material constant: see theory of contact mechanics [Pa]
    gamma : float, optional
        constant in contact mechanics. The default is 1. see theory of contact mechanics
    eps : float, optional
        constant in contact mechanics. The default is 0.7493. see theory of contact mechanics

    Returns
    -------
    u : ufloat
        average surface separation [m]

    '''
    u = gamma * alpha(q,H) * h_rms * unp.log(beta(q, H)*eps*q_0*h_rms*E/p)
    return u


''' Piezo and membrane parameters '''
Y_steel = ufloat(195e9, 0.01*195e9)
nu_steel = ufloat(0.30, 0.30*0.05)
Y_p = ufloat(67e9, 0.01*67e9)
d_31 = ufloat(2.1e-10, 0.01*2.1e-10)
T_d = ufloat(100e-6, 10e-6)
R_m = ufloat(10.0e-3, 10.0e-3*0.1)
T_p = ufloat(200e-6, 20e-6)
R_p = ufloat(8.0e-3, 8.0e-3*0.1)
C_p, C_E = _Cp(Y_steel, nu_steel, Y_p, T_d, R_m, T_p, R_p), _CEs(Y_steel, nu_steel, Y_p, d_31, T_d, R_m, T_p, R_p)
U_2, U_1 = ufloat(400, 0.8), ufloat(-80, 0.16) # Voltages of the piezo

A_piezo, A_contact = ufloat(254.47, 254.47*0.1), ufloat(239.63, 0.1*239.63)
Y_parylene = ufloat(5e9, 0.5e9)
nu_parylene = ufloat(0.3, 0.1)

''' Values for Contact Mechanis '''
H = 3-ufloat(2.014, 0.032)
h_q = ufloat(1.119e-6, 0.174e-6)     # Excel Tabelle
L, a = 1e-3, 1e-3/2500
q_0, q_1 = 2*np.pi/L, 2*np.pi/a
E = ((1-nu_steel**2)/Y_steel + (1-nu_parylene**2)/Y_parylene)**(-1) # steel-parylene contact; must be modifies for other contact materials

''' Values for the leakage calculation '''
k, T = sc.Boltzmann, ufloat(303, 303*0.02)
w = ufloat(2.2e-3, 0.01e-3)
L_ges = 1570e-6
L_ringe = ufloat(0.6e-3, 0.05e-3)
L = L_ges - L_ringe
mu = ufloat(0.7978e-3, 0.7978e-3*0.01)
delta_p = np.array([0,25, 50, 75, 100, 150, 200, 300, 400, 500, 600]) * 100
p_err = 0.01*(delta_p + 101325)
delta_p = unp.uarray(delta_p, p_err)

''' Calculate the squeezing pressure, average surface separation, the slip length and the resulting leakage '''
p = p_block(U_2, U_1, C_E, C_p) * A_piezo/A_contact
h = average_separation(p, h_q, q_1/q_0, q_0, H, E)

Q = h**3*w / (12*mu) * delta_p / L  * 60e9
np.savetxt('V32002-Typ2_400V.txt', Q, fmt='%r')



